/*
===============================================================================

	FILE:	codec.cpp
	
	CONTENTS:
	
    see corresponding header file
	
	PROGRAMMERS:
	
		martin isenburg@cs.unc.edu
	
	COPYRIGHT:
	
		(c) 2002 martin isenburg@cs.unc.edu
		
		This software is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	
	CHANGE HISTORY:
	
    see corresponding header file
	
===============================================================================
*/
#include "codec.h"

#include "arithmeticmodel.h"
#include "arithmeticencoder.h"
#include "arithmeticdecoder.h"

CFcodec::CFcodec()
{
	ae_conn = 0;
	ae_geom = 0;
	ad = 0;
}

CFcodec::~CFcodec()
{
}

void CFcodec::initEnc(int boundary_degrees, int* has_boundary_degree, int interior_degrees, int* has_interior_degree, int pos_bits)
{
	int i;
	int table[2] = {100,1};

	if (ae_conn)
	{
		delete ae_conn;
	}
	ae_conn = new ArithmeticEncoder();

	if (ae_geom)
	{
		delete ae_geom;
	}
	ae_geom = new ArithmeticEncoder();

	be = 0;
	ie = 0;

	ib_on = 0;
	ib_off = 0;
	av_on = 0;
	av_off = 0;

	boundary_degree_dist = new int[boundary_degrees];
	interior_degree_dist = new int[interior_degrees];

	// saving bits  that indicate which boundary edge degrees are there
	for (i = 0; i < boundary_degrees; i++)
	{
		boundary_degree_dist[i] = 0;
		if (has_boundary_degree[i])
		{
			ae_conn->encode(2, 1);
		}
		else
		{
			ae_conn->encode(2, 0);
		}
	}
	// saving bits that indicate which interior edge degrees are there
	for (i = 0; i < interior_degrees; i++)
	{
		interior_degree_dist[i] = 0;
		if (has_interior_degree[i])
		{
			ae_conn->encode(2, 1);
		}
		else
		{
			ae_conn->encode(2, 0);
		}
	}

	// connectivity

	amBoundaryEdgeDegree = new ArithmeticModel(boundary_degrees, true, has_boundary_degree); // 0: deg2, 1: deg3, 2: deg4, 3: deg5 ...
	amLargeBoundaryEdgeDegree = new ArithmeticModel(8,true,0); // larger
	amInteriorEdgeDegree = new ArithmeticModel(interior_degrees, true, has_interior_degree); // 0: deg2, 1: deg3, 2: deg4, 3: deg5 ...
	amLargeInteriorEdgeDegree = new ArithmeticModel(8,true,0); // larger
	amIsBoundary = new ArithmeticModel(2,true,table); // 0: no, 1: yes
	amAlreadyVisited = new ArithmeticModel(2,true,table); // 0: no, 1: yes

	// geometry

	amSign = new ArithmeticModel(2,true,0);
	amSignOther = new ArithmeticModel(2,true,0);

	if (pos_bits == 18)
	{
		// 7-4-3-2-1 = 17+1 bits
		bitsUsed[0] = 7;
		bitsUsed[1] = 4;
		bitsUsed[2] = 3;
		bitsUsed[3] = 2;
		bitsUsed[4] = 1;
	}
	else if (pos_bits == 16)
	{
		// 6-3-3-2-1 = 15+1 bits
		bitsUsed[0] = 6;
		bitsUsed[1] = 3;
		bitsUsed[2] = 3;
		bitsUsed[3] = 2;
		bitsUsed[4] = 1;
	}
	else if (pos_bits == 14)
	{
		// 5-3-2-2-1 = 13+1 bits
		bitsUsed[0] = 5;
		bitsUsed[1] = 3;
		bitsUsed[2] = 2;
		bitsUsed[3] = 2;
		bitsUsed[4] = 1;
	}
	else if (pos_bits == 12)
	{
		// 5-3-2-1 = 11+1 bits
		bitsUsed[0] = 5;
		bitsUsed[1] = 3;
		bitsUsed[2] = 2;
		bitsUsed[3] = 1;
		bitsUsed[4] = 0;
	}
	else if (pos_bits == 10)
	{
		// 5-3-1 = 9+1 bits
		bitsUsed[0] = 5;
		bitsUsed[1] = 3;
		bitsUsed[2] = 1;
		bitsUsed[3] = 0;
		bitsUsed[4] = 0;
	}
	else if (pos_bits == 8)
	{
		// 4-2-1 = 7+1 bits
		bitsUsed[0] = 4;
		bitsUsed[1] = 2;
		bitsUsed[2] = 1;
		bitsUsed[3] = 0;
		bitsUsed[4] = 0;
	}
	for (i = 0; i < 5; i++)
	{
		if (bitsUsed[i])
		{
			bitsMask[i] = (1 << bitsUsed[i]);
			amBits[i] = new ArithmeticModel(bitsMask[i],true,0);
			amBitsOther[i] = new ArithmeticModel(bitsMask[i],true,0);
			bitsMask[i] -= 1;
		}
		else
		{
			amBits[i] = 0;
      amBitsOther[i] = 0;
		}
	}
}

void CFcodec::initDec()
{
}

void CFcodec::doneEnc()
{
	ae_conn->done();
	ae_geom->done();

	delete amBoundaryEdgeDegree;
	delete amLargeBoundaryEdgeDegree;
	delete amInteriorEdgeDegree;
	delete amLargeInteriorEdgeDegree;
	delete amIsBoundary;
	delete amAlreadyVisited;

	delete amSign;
	for (int i = 0; i < 5; i++)
	{
		if (amBits[i]) delete amBits[i];
	}
}

void CFcodec::doneDec()
{
}

void CFcodec::writeBoundaryEdgeDegree(int degree)
{
	be++;

	if (degree > amBoundaryEdgeDegree->nsym)
	{
		boundary_degree_dist[amBoundaryEdgeDegree->nsym-1]++;
		ae_conn->encode(amBoundaryEdgeDegree, amBoundaryEdgeDegree->nsym-1);
		writeLargeBoundaryEdgeDegree(degree - (amBoundaryEdgeDegree->nsym-1));
	}
	else
	{
		boundary_degree_dist[degree-2]++;
		ae_conn->encode(amBoundaryEdgeDegree, degree-2);
	}
}

void CFcodec::writeLargeBoundaryEdgeDegree(int degree)
{
	while (degree > (amLargeBoundaryEdgeDegree->nsym))
	{
		ae_conn->encode(amLargeBoundaryEdgeDegree, amLargeBoundaryEdgeDegree->nsym-1);
		degree -= (amLargeBoundaryEdgeDegree->nsym-1);
	}
	ae_conn->encode(amLargeBoundaryEdgeDegree, degree-2);
}

void CFcodec::writeInteriorEdgeDegree(int degree)
{
	ie++;

	if (degree > amInteriorEdgeDegree->nsym)
	{
		interior_degree_dist[amInteriorEdgeDegree->nsym-1]++;
		ae_conn->encode(amInteriorEdgeDegree, amInteriorEdgeDegree->nsym-1);
		writeLargeInteriorEdgeDegree(degree - (amInteriorEdgeDegree->nsym-1));
	}
	else
	{
		interior_degree_dist[degree-2]++;
		ae_conn->encode(amInteriorEdgeDegree, degree-2);
	}
}

void CFcodec::writeLargeInteriorEdgeDegree(int degree)
{
	while (degree > amLargeInteriorEdgeDegree->nsym)
	{
		ae_conn->encode(amLargeInteriorEdgeDegree, amLargeInteriorEdgeDegree->nsym-1);
		degree -= (amLargeInteriorEdgeDegree->nsym-1);
	}
	ae_conn->encode(amLargeInteriorEdgeDegree, degree-2);
}

void CFcodec::writeIsBoundary(int truefalse)
{
	if (truefalse)
	{
		ib_on++;
	}
	else
	{
		ib_off++;
	}
	ae_conn->encode(amIsBoundary, truefalse);
}

void CFcodec::writeAlreadyVisited(int truefalse)
{
	if (truefalse)
	{
		av_on++;
	}
	else
	{
		av_off++;
	}
	ae_conn->encode(amAlreadyVisited, truefalse);
}

void CFcodec::writeRange(int range, int index)
{
	ae_conn->encode(range, index);
}

void CFcodec::writePositionCorrector(int* corr)
{
	int i, j, c;
	int sign, bits[5];

	for (i = 0; i < 3; i++)
	{
		c = corr[i];
    if (c > 0)			// positive
		{
			sign = 0;
		}
		else if (c < 0)	// negative
		{
			sign = 1;
			c = -c;
		}
		else						// zero
		{
			sign = -1;
		}
		for (j = 0; j < 5; j++)
		{
			if (amBits[j])
			{
				bits[j] = c & bitsMask[j];
				c = c >> bitsUsed[j];
			}
		}
		if (c)
		{
//			cout << "FATAL ERROR: c is not zero: " << c << endl;
			return;
		}
		for (j = 4; j >= 0; j--)
		if (amBits[j])
		{
			ae_geom->encode(amBits[j], bits[j]);
		}
    if (sign != -1) // positive or negative (e.g. not null)
		{
			ae_geom->encode(amSign, sign);
		}
	}
}

void CFcodec::writePositionCorrectorOther(int* corr)
{
	int i, j, c;
	int sign, bits[5];

	for (i = 0; i < 3; i++)
	{
		c = corr[i];
    if (c > 0)			// positive
		{
			sign = 0;
		}
		else if (c < 0)	// negative
		{
			sign = 1;
			c = -c;
		}
		else						// zero
		{
			sign = -1;
		}
		for (j = 0; j < 5; j++)
		{
			if (amBitsOther[j])
			{
				bits[j] = c & bitsMask[j];
				c = c >> bitsUsed[j];
			}
		}
		if (c)
		{
//			cout << "FATAL ERROR: c is not zero: " << c << endl;
			return;
		}
		for (j = 4; j >= 0; j--)
		if (amBitsOther[j])
		{
			ae_geom->encode(amBitsOther[j], bits[j]);
		}
    if (sign != -1) // positive or negative (e.g. not null)
		{
			ae_geom->encode(amSignOther, sign);
		}
	}
}
